from __future__ import annotations

import collections
import contextlib
import importlib
import types
import typing

import typing_extensions

from freedom import util
from freedom.domain import command as command_
from freedom.domain import command_handler as command_handler_
from freedom.domain import event as event_
from freedom.domain import event_handler as event_handler_
from freedom.domain import repository as repository_

if typing.TYPE_CHECKING:
    from freedom.application import command_bus as command_bus_
    from freedom.application import event_emitter as event_emitter_
    from freedom.application import provider

_CallableT = typing.TypeVar("_CallableT", bound=typing.Callable[..., typing.Any])
_ResultT = typing.TypeVar("_ResultT", bound=command_handler_.CommandResult)


def collect_domain_events(
    result: _ResultT,
    kwargs: typing.Mapping[str, typing.Any],
) -> _ResultT:
    repositories = typing.cast(
        typing.Iterable[repository_.AnyRepository],
        filter(lambda x: isinstance(x, repository_.Repository), kwargs.values()),
    )
    domain_events = []
    for repository in repositories:
        domain_events.extend(repository.collect_events())

    result.events.extend(domain_events)
    return result


class TransactionContext:
    def __init__(
        self,
        application: Application,
        command_bus: command_bus_.CommandBus,
        dependency_provider: typing.Optional[provider.AnyDependencyProviderType] = None,
        event_emitter: typing.Optional[event_emitter_.EventEmitter] = None,
    ) -> None:
        self._application = application
        self._event_emitter = event_emitter
        self._command_bus = command_bus
        self._dependency_provider = dependency_provider

        # ... declaration explanation
        self._executing_task: bool = False

    def __enter__(self) -> typing_extensions.Self:
        self._application._on_enter_transaction_context(self)
        return self

    @typing.overload
    def __exit__(self, exc_type: None, exc_val: None, exc_tb: None) -> None: ...

    @typing.overload
    def __exit__(
        self,
        exc_type: typing.Type[BaseException],
        exc_val: BaseException,
        exc_tb: types.TracebackType,
    ) -> None: ...

    @typing.no_type_check
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_val: typing.Optional[BaseException],
        exc_tb: typing.Optional[types.TracebackType],
    ) -> None:
        self._application._on_exit_transaction_context(self, exc_type, exc_val, exc_tb)

    async def execute_command(
        self, command: command_.Command
    ) -> command_handler_.CommandResult:
        with self._lock_transaction():
            command_result = await self._command_bus.execute(command)

            event_queue = command_result.events.copy()
            while len(event_queue) > 0:
                event = event_queue.pop(0)

                if isinstance(event, event_.Event):
                    # Чтобы в дальнейшем реализовать интеграционные ивенты in-box & out-box pattern
                    await self.handle_domain_event(event)
                    # event_results = await self.handle_domain_event(event)
                    # event_queue.extend(event_results.events)

            if command_result.is_success():
                result = command_handler_.CommandResult.success(
                    entity_id=command_result.entity_id,
                    events=command_result.events,
                )
            else:
                # FIXME: пока что берем только первую ошибку
                err = command_result.errors[0]
                result = command_handler_.CommandResult.failure(
                    fail_message=err.message,
                    exception=err.exception,
                    exception_info=err.exception_info,
                )

            return result

    async def handle_domain_event(self, event: event_.Event) -> None:
        if self._event_emitter is None:
            raise ValueError(
                "Event emitter is not set, when event sourced aggregate is used."
            )
        await self._event_emitter.emit(event)

    def get_service(self, service_cls: typing.Any) -> typing.Any:
        if self._dependency_provider is None:
            raise ValueError("Application does not have dependency provider.")
        return self._dependency_provider.get_dependency(service_cls)

    @contextlib.contextmanager
    def _lock_transaction(self) -> typing.Iterator[None]:
        if self._executing_task:
            raise RuntimeError(
                "Cannot execute command while another task is being executed."
            )

        self._executing_task = True
        yield
        self._executing_task = False


class ApplicationModule:
    __slots__: typing.Sequence[str] = (
        "_name",
        "_version",
        "_command_handlers",
        "_event_handlers",
        "_query_handlers",
    )

    def __init__(self, name: str, version: float) -> None:
        self._name = name
        self._version = version
        self._command_handlers: typing.Dict[
            typing.Type[command_.Command],
            command_handler_.AnyCommandHandlerType,
        ] = {}
        self._event_handlers: typing.Dict[
            typing.Type[event_.Event], typing.List[event_handler_.AnyEventHandlerType]
        ] = collections.defaultdict(list)

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> float:
        return self._version

    def load(self, path: str) -> None:
        importlib.import_module(path)

    def with_command_handler(
        self,
        handler: command_handler_.AnyCommandHandlerType,
        command_cls: typing.Optional[command_.CommandType] = None,
    ) -> typing_extensions.Self:
        if (command := command_cls) is None:
            command, *_ = util.get_type_hints(
                handler.__init__,
                subclass_of=(command_handler_.CommandHandler,),
            ).values()

        self._command_handlers[command] = handler
        return self

    def with_event_handler(
        self,
        handler: event_handler_.AnyEventHandlerType,
        event_cls: typing.Optional[event_.EventType] = None,
    ) -> typing_extensions.Self:
        if (event := event_cls) is None:
            event, *_ = util.get_type_hints(
                handler.__init__,
                subclass_of=(event_handler_.EventHandler,),
            ).values()

        self._event_handlers[event].append(handler)
        return self

    def event_handler(
        self,
        event_cls: typing.Optional[event_.EventType] = None,
    ) -> typing.Callable[
        [event_handler_.AnyEventHandlerTypeT], event_handler_.AnyEventHandlerTypeT
    ]:
        def decorator(
            handler: event_handler_.AnyEventHandlerTypeT,
        ) -> event_handler_.AnyEventHandlerTypeT:
            self.with_event_handler(handler, event_cls)
            return handler

        return decorator

    def command_handler(
        self,
        command_cls: typing.Optional[command_.CommandType] = None,
    ) -> typing.Callable[
        [command_handler_.AnyCommandHandlerTypeT],
        command_handler_.AnyCommandHandlerTypeT,
    ]:
        def decorator(
            handler: command_handler_.AnyCommandHandlerTypeT,
        ) -> command_handler_.AnyCommandHandlerTypeT:
            self.with_command_handler(handler, command_cls)
            return handler

        return decorator


class Application(ApplicationModule):
    def __init__(
        self,
        name: str,
        version: float,
        *,
        command_bus: command_bus_.CommandBus,
        dependency_provider: typing.Optional[provider.AnyDependencyProviderType] = None,
        event_emitter: typing.Optional[event_emitter_.EventEmitter] = None,
    ) -> None:
        super().__init__(name, version)
        self._command_bus = command_bus
        self._event_emitter = event_emitter
        self._dependency_provider = dependency_provider
        self._on_enter_transaction_context: typing.Callable[
            [TransactionContext], None
        ] = lambda ctx: None
        self._on_exit_transaction_context: typing.Callable[..., None] = (
            lambda ctx, exc_type, exc_val, exc_tb: None
        )
        self._modules: typing.Set[ApplicationModule] = {self}

    def include_module(self, module: ApplicationModule) -> None:
        if not isinstance(module, ApplicationModule):
            raise TypeError("Can only include ApplicationModule instances")

        # FIXME: event handlers, ... , etc
        for command, command_handler in module._command_handlers.items():
            self._command_bus.subscribe(command_handler, command)

        if self._event_emitter is not None:
            for event, event_handlers in module._event_handlers.items():
                for handler in event_handlers:
                    self._event_emitter.subscribe(handler, event)

        self._modules.add(module)

    def on_enter_transaction_context(self, callable_: _CallableT) -> _CallableT:
        self._on_enter_transaction_context = callable_
        return callable_

    def on_exit_transaction_context(self, callable_: _CallableT) -> _CallableT:
        self._on_exit_transaction_context = callable_
        return callable_

    def get_command_handler(
        self, command: command_.Command
    ) -> command_handler_.AnyCommandHandlerType:
        command_cls = type(command)
        for app_module in self._modules:
            handler_func = app_module._command_handlers.get(command_cls)
            if handler_func is not None:
                return handler_func

        raise KeyError(f"No command handler found for command {command_cls}")

    def get_event_handlers(
        self, event: event_.Event
    ) -> typing.List[event_handler_.AnyEventHandlerType]:
        event_cls = type(event)
        event_handlers = []
        for app_module in self._modules:
            event_handlers.extend(app_module._event_handlers.get(event_cls, []))

        return event_handlers

    def transaction_context(self) -> TransactionContext:
        ctx = TransactionContext(
            self,
            command_bus=self._command_bus,
            event_emitter=self._event_emitter,
            dependency_provider=self._dependency_provider,
        )
        return ctx

    async def execute_command(
        self, command: command_.Command
    ) -> command_handler_.CommandResult:
        with self.transaction_context() as ctx:
            return await ctx.execute_command(command)



from __future__ import annotations

import asyncio
import types
import typing
import weakref

import typing_extensions as typingext

from freedom.domain import event as event_

if typing.TYPE_CHECKING:
    from freedom.application import event_emitter as emitter_


def _generate_weak_listener(
    reference: weakref.WeakMethod[typing.Any],
) -> typing.Callable[[event_.Event], typing.Coroutine[typing.Any, typing.Any, None]]:
    async def call_weak_method(event: event_.Event) -> None:
        method = reference()
        if method is None:
            raise TypeError(
                "dead weak referenced subscriber method cannot be executed, "
                "try actually closing your event streamers"
            )

        await method(event)

    return call_weak_method


class EventStream:
    __slots__: typing.Sequence[str] = (
        "__weakref__",
        "_active",
        "_event",
        "_event_manager",
        "_event_type",
        "_filters",
        "_limit",
        "_queue",
        "_registered_listener",
        "_timeout",
    )

    __weakref__: typing.Optional[weakref.ref[EventStream]]

    def __init__(
        self,
        emitter: emitter_.EventEmitter,
        event_type: typing.Type[event_.Event],
        *,
        timeout: typing.Union[float, int, None],
        limit: typing.Optional[int] = None,
    ) -> None:
        self._active = False
        self._event: typing.Optional[asyncio.Event] = None
        self._event_manager = emitter
        self._event_type = event_type
        self._limit = limit
        self._queue: typing.List[event_.Event] = []
        self._registered_listener: typing.Optional[
            typing.Callable[
                [event_.Event], typing.Coroutine[typing.Any, typing.Any, None]
            ]
        ] = None
        self._timeout = timeout

    def __enter__(self) -> typingext.Self:
        self.open()
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_val: typing.Optional[BaseException],
        exc_tb: typing.Optional[types.TracebackType],
    ) -> None:
        self.close()

    def __aiter__(self) -> typing.AsyncIterator[event_.Event]:
        return self

    async def __anext__(self) -> event_.Event:
        if not self._active:
            raise TypeError("stream must be started with before entering it")

        while not self._queue:
            if not self._event:
                self._event = asyncio.Event()

            try:
                await asyncio.wait_for(self._event.wait(), timeout=self._timeout)
            except asyncio.TimeoutError:
                raise StopAsyncIteration from None

            self._event.clear()

        return self._queue.pop(0)

    def __await__(self) -> typing.Generator[None, None, typing.Sequence[event_.Event]]:
        async def _await_all() -> typing.Sequence[event_.Event]:
            self.open()
            result = [
                event
                async for event in typing.cast(typing.AsyncIterable[event_.Event], self)
            ]
            self.close()
            return result

        return _await_all().__await__()

    def __del__(self) -> None:
        if self._active:
            # _LOGGER.warning(
            # "active %r streamer fell out of scope before being closed",
            # self._event_type.__name__)
            self.close()

    async def _listener(self, event: event_.Event) -> None:
        if self._limit is not None and len(self._queue) >= self._limit:
            return

        self._queue.append(event)
        if self._event:
            self._event.set()

    def close(self) -> None:
        if self._active and self._registered_listener is not None:
            try:
                self._event_manager.unsubscribe(
                    self._registered_listener, self._event_type
                )
            except ValueError as exc:
                # FIXME
                pass

            self._registered_listener = None

        self._active = False

    def filter(
        self, condition: typing.Callable[[event_.Event], bool]
    ) -> typingext.Self:
        if self._active:
            self._queue = [entry for entry in self._queue if condition(entry)]

        return self

    def open(self) -> None:
        if not self._active:
            reference = weakref.WeakMethod(self._listener)
            listener = _generate_weak_listener(reference)

            self._registered_listener = listener
            self._event_manager.subscribe(listener, self._event_type)
            self._active = True



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "DataMapper",
    "MapperModelT",
    "MapperEntityT",
    "AnyDataMapper",
    "AnyDataMapperType",
)

import abc
import typing

import typing_extensions as typingext

AnyDataMapper: typingext.TypeAlias = "DataMapper[typing.Any, typing.Any]"
AnyDataMapperType: typingext.TypeAlias = typing.Type["AnyDataMapper"]

MapperEntityT = typing.TypeVar("MapperEntityT")
MapperModelT = typing.TypeVar("MapperModelT")


class DataMapper(typing.Generic[MapperEntityT, MapperModelT], abc.ABC):
    @abc.abstractmethod
    def model_to_entity(self, instance: MapperModelT) -> MapperEntityT: ...

    @abc.abstractmethod
    def entity_to_model(self, entity: MapperEntityT) -> MapperModelT: ...



from __future__ import annotations

__all__: typing.Sequence[str] = ("Uuid4EntityId",)

import typing
import uuid

from freedom.domain import entity_id


class Uuid4EntityId(entity_id.EntityId):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def next_id(cls) -> Uuid4EntityId:
        self = cls(int=uuid.uuid4().int)
        return self



from __future__ import annotations

__all__: typing.Sequence[str] = ("EventEmitter",)

import abc
import asyncio
import typing

if typing.TYPE_CHECKING:
    from freedom.application import inflector
    from freedom.domain import event as event_
    from freedom.domain import event_handler as handler_


class EventEmitter(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def emit(self, event: event_.Event, /) -> asyncio.Future[typing.Any]: ...

    @abc.abstractmethod
    def listen(
        self,
        handler: typing.Optional[handler_.AnyEventHandlerTypeT] = None,
        event: typing.Optional[event_.EventType] = None,
    ) -> typing.Union[
        handler_.AnyEventHandlerTypeT,
        typing.Callable[
            [handler_.AnyEventHandlerTypeT],
            handler_.AnyEventHandlerTypeT,
        ],
    ]: ...

    @abc.abstractmethod
    def get_handler_for(
        self,
        event: event_.EventType,
    ) -> typing.Optional[handler_.AnyEventHandlerType]: ...

    @abc.abstractmethod
    def get_handlers(self) -> inflector.TargetHandlersViewType[
        event_.EventType,
        handler_.AnyEventHandlerType,
    ]: ...

    @abc.abstractmethod
    def subscribe(
        self,
        handler: handler_.AnyEventHandlerType,
        event: event_.EventType,
    ) -> None: ...

    @abc.abstractmethod
    def unsubscribe(
        self, handler: handler_.AnyEventHandlerType, event: event_.EventType
    ) -> None: ...

    @abc.abstractmethod
    async def wait_for(
        self,
        event_type: event_.EventType,
        /,
        timeout: typing.Union[float, int, None],
        predicate: typing.Optional[typing.Callable[[event_.EventType], bool]] = None,
    ) -> event_.Event: ...

    # @abc.abstractmethod
    # def attach_stream(self, stream):
    #     pass



from __future__ import annotations

import typing

from freedom.domain.valueobject import ValueObject

if typing.TYPE_CHECKING:
    from freedom.domain import event_handler as handler_

EventType = typing.Type["Event"]


class Event(ValueObject):
    pass


class ExceptionEvent(Event):
    __slots__: typing.Sequence[str] = (
        "exception",
        "failed_event",
        "failed_callback",
    )

    def __init__(
        self,
        exception: BaseException,
        failed_event: Event,
        failed_callback: handler_.AnyEventHandler,
    ) -> None:
        self.exception = exception
        self.failed_event = failed_event
        self.failed_callback = failed_callback

    async def retry(self) -> None:
        await self.failed_callback.handle(self.failed_event)



from __future__ import annotations

import inspect
import typing

from freedom import sentinel
from freedom.application import provider


class InMemoryDependencyProvider(
    provider.DependencyProvider[typing.Type[typing.Any], typing.Any]
):
    __slots__: typing.Sequence[str] = ("_dependencies",)

    def __init__(
        self,
        overrides: typing.Optional[
            typing.Mapping[typing.Type[typing.Any], typing.Any]
        ] = None,
    ) -> None:
        self._dependencies: typing.Dict[typing.Type[typing.Any], typing.Any] = {}

        if overrides is not None:
            for k, v in overrides.items():
                self.register_dependency(k, v)

    def register_dependency(
        self,
        identifier: typing.Type[typing.Any],
        dependency: typing.Any,
    ) -> None:
        if not inspect.isclass(identifier):
            raise ValueError("Dependency identifier must be a type.")
        self._dependencies[identifier] = dependency

    def get_dependency(self, identifier: typing.Type[typing.Any]) -> typing.Any:
        dependency = self._dependencies.get(identifier, sentinel.NOTHING)
        if dependency is sentinel.NOTHING:
            raise ValueError(f"No dependency found with identifier {identifier!r}.")
        return dependency



from __future__ import annotations

import abc
import typing
import uuid

from freedom.domain.valueobject import ValueObject


class EntityId(ValueObject, abc.ABC):
    __slots__: typing.Sequence[str] = ()

    def __init__(self, int: int) -> None:
        self._int = int

    @classmethod
    @abc.abstractmethod
    def next_id(cls, **kwargs: typing.Any) -> EntityId: ...

    @property
    def int(self) -> int:
        return self._int

    @property
    def str(self) -> str:
        return str(self.int)


class EntityIdUuid4(EntityId):
    @classmethod
    def next_id(cls, **kwargs: typing.Any) -> EntityIdUuid4:
        return cls(int=uuid.uuid4().int)


class EntityIdSequential(EntityId):
    @classmethod
    def next_id(cls, **kwargs: typing.Any) -> EntityIdSequential:
        next_id = kwargs.pop("next_id")
        if not isinstance(next_id, int):
            next_id = int(next_id)
        return cls(int=next_id)



from __future__ import annotations

import typing

from freedom.domain.business_rule import BusinessRuleValidationMixin
from freedom.domain.entity_id import EntityId

_EntityIdT = typing.TypeVar("_EntityIdT", bound=EntityId)


class Entity(BusinessRuleValidationMixin, typing.Generic[_EntityIdT]):
    __slots__: typing.Sequence[str] = ("_id",)

    def __init__(self, id: _EntityIdT) -> None:
        self._id = id

    @property
    def id(self) -> _EntityIdT:
        return self._id



from __future__ import annotations

import abc
import typing

import typing_extensions as typingext

MiddlewareSigType: typingext.TypeAlias = typing.Callable[
    [typing.Any, typing.Any], typing.Awaitable[typing.Any]
]
PartialMiddlewareSigType: typingext.TypeAlias = typing.Callable[
    [typing.Any], typing.Awaitable[typing.Any]
]


# TODO: Make middlewares Generic ?


class Middleware(abc.ABC):
    @abc.abstractmethod
    async def __call__(
        self,
        message: typing.Any,
        next_middleware: PartialMiddlewareSigType,
    ) -> typing.Any: ...


class MiddlewareChain(abc.ABC):
    @abc.abstractmethod
    async def __call__(self, message: typing.Any) -> typing.Any: ...

    @abc.abstractmethod
    def set_next(
        self,
        middleware: MiddlewareSigType,
        next_middleware: MiddlewareSigType,
    ) -> PartialMiddlewareSigType: ...



from __future__ import annotations

import abc
import types
import typing

import typing_extensions

from freedom.domain import event as event_

_EventT = typing.TypeVar("_EventT")


class EventStream(typing.Generic[_EventT]):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def __enter__(self) -> typing_extensions.Self: ...

    @typing.overload
    @abc.abstractmethod
    def __exit__(self, exc_type: None, exc_val: None, exc_tb: None) -> None: ...

    @typing.overload
    @abc.abstractmethod
    def __exit__(
        self,
        exc_type: typing.Type[BaseException],
        exc_val: BaseException,
        exc_tb: types.TracebackType,
    ) -> None: ...

    @abc.abstractmethod
    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_val: typing.Optional[BaseException],
        exc_tb: typing.Optional[types.TracebackType],
    ) -> None: ...

    @abc.abstractmethod
    def __aiter__(self) -> typing_extensions.Self: ...

    @abc.abstractmethod
    async def __anext__(self): ...

    @abc.abstractmethod
    def __await__(
        self,
    ) -> typing.Generator[None, None, typing.Sequence[event_.Event]]: ...

    @abc.abstractmethod
    def __del__(self) -> None: ...

    @abc.abstractmethod
    def close(self) -> None: ...

    @abc.abstractmethod
    def filter(
        self, filter_: typing.Callable[[typing.Sequence[event_.Event]], bool], /
    ): ...

    @abc.abstractmethod
    def open(self) -> None: ...



from __future__ import annotations

__all__: typing.Sequence[str] = ("InMemoryEventEmitter",)

import asyncio
import collections
import inspect
import typing

from freedom import util
from freedom.application import event_emitter
from freedom.application import inflector as inflector_
from freedom.application import provider as provider_
from freedom.domain import event as event_
from freedom.domain import event_handler as handler_
from freedom.infrastructure import inflector as inflector_impl


class InMemoryEventEmitter(event_emitter.EventEmitter):
    __slots__: typing.Sequence[str] = (
        "_waiters",
        "_inflector",
        "_provider",
        "_dispatch_on_exc",
    )

    def __init__(
        self,
        *,
        provider: typing.Optional[provider_.AnyDependencyProviderType] = None,
        inflector: typing.Optional[
            inflector_.Inflector[
                event_.EventType,
                typing.List[handler_.AnyEventHandlerType],
            ]
        ] = None,
        dispatch_on_exc: bool = True,
    ) -> None:
        if inflector is None:
            inflector = typing.cast(
                inflector_.Inflector[
                    event_.EventType,
                    typing.List[handler_.AnyEventHandlerType],
                ],
                inflector_impl.InMemoryInflector(
                    handler_type=handler_.EventHandler,
                ),
            )

        self._dispatch_on_exc = dispatch_on_exc
        self._waiters: typing.Any = collections.defaultdict(set)  # FIXME
        self._inflector = inflector
        self._provider = provider

    def emit(self, event: event_.Event, /) -> asyncio.Future[typing.Any]:
        event_type = type(event)
        tasks: typing.List[typing.Coroutine[None, typing.Any, None]] = []

        handlers = self.get_handler_for(event_type)
        if handlers:
            for handler in handlers:
                if inspect.isclass(handler):
                    dependencies: typing.Dict[str, typing.Any] = {}
                    if self._provider is not None:
                        dependencies.update(
                            self._provider.get_dependencies(handler.__init__)
                        )

                    handler = handler(**dependencies)

                tasks.append(self._trigger_handler(handler, event))

        waiter_set = self._waiters[event_type]

        for waiter in tuple(waiter_set):
            predicate, future = waiter
            if not future.done():
                try:
                    if predicate and not predicate(event):
                        continue
                except Exception as exc:
                    future.set_exception(exc)
                else:
                    future.set_result(event)

            waiter_set.remove(waiter)

        if tasks:
            return asyncio.gather(*tasks)

        return util.as_completed_future()

    def listen(
        self,
        handler: typing.Optional[handler_.AnyEventHandlerType] = None,
        event: typing.Optional[event_.EventType] = None,
    ) -> typing.Union[
        handler_.AnyEventHandlerType,
        typing.Callable[
            [handler_.AnyEventHandlerType],
            handler_.AnyEventHandlerType,
        ],
    ]:
        return self._inflector.listen(handler, event)

    def get_handler_for(
        self,
        event: event_.EventType,
    ) -> typing.Optional[typing.List[handler_.AnyEventHandlerType]]:
        return self._inflector.get_handler_for(event)

    def get_handlers(self) -> inflector_.TargetHandlersViewType[
        event_.EventType,
        typing.List[handler_.AnyEventHandlerType],
    ]:
        return self._inflector.get_handlers()

    def subscribe(
        self,
        handler: handler_.AnyEventHandlerType,
        event: event_.EventType,
    ) -> None:
        handlers = self.get_handler_for(event)
        if handlers is None:
            handlers = []

        handlers.append(handler)
        self._inflector.subscribe(handlers, event, allow_many=True)

    def unsubscribe(
        self, handler: handler_.AnyEventHandlerType, event: event_.EventType
    ) -> None:
        self._inflector.unsubscribe(event, handler)

    async def wait_for(
        self,
        event_type: event_.EventType,
        /,
        timeout: typing.Union[float, int, None],
        predicate: typing.Optional[typing.Callable[[typing.Any], bool]] = None,
    ) -> event_.Event:
        future = util.get_loop().create_future()

        if predicate is None:
            predicate = lambda _: True

        waiter_set = self._waiters[event_type]
        item = (predicate, future)
        waiter_set.add(item)

        try:
            event = await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError:
            waiter_set.remove(item)
            raise

        return typing.cast(event_.Event, event)

    async def _trigger_handler(
        self,
        callback: typing.Callable[..., typing.Awaitable[None]],
        event: event_.Event,
    ) -> None:
        try:
            if isinstance(callback, handler_.EventHandler):
                await callback.handle(event)
                return
            await callback(event)
        except Exception as exc:
            if not self._dispatch_on_exc:
                raise exc

            exc_event = event_.ExceptionEvent(
                exception=exc,
                failed_event=event,
                failed_callback=callback,
            )
            await self.emit(exc_event)



from __future__ import annotations

import inspect
import types
import typing

from freedom import util
from freedom.application import inflector

_AnyHandlerT = typing.TypeVar("_AnyHandlerT")
_TargetHandlerT = typing.TypeVar("_TargetHandlerT", bound=typing.Type[typing.Any])
_TargetT = typing.TypeVar("_TargetT", bound=typing.Type[typing.Any])


class InMemoryInflector(inflector.Inflector[_TargetT, _TargetHandlerT]):
    __slots__ = (
        "_handlers",
        "_handler_type",
    )

    def __init__(
        self,
        *,
        handler_type: _TargetHandlerT,  # To get args from generic[]
        handlers: typing.Optional[
            inflector.TargetHandlersType[_TargetT, _TargetHandlerT]
        ] = None,
    ) -> None:
        if handlers is None:
            handlers = {}

        self._handlers = handlers
        self._handler_type = handler_type

    def listen(
        self,
        handler: typing.Optional[_AnyHandlerT] = None,
        target: typing.Optional[_TargetT] = None,
    ) -> typing.Union[_AnyHandlerT, typing.Callable[[_AnyHandlerT], _AnyHandlerT]]:
        def wrap(target_handler: _AnyHandlerT) -> _AnyHandlerT:
            if not inspect.isclass(target_handler):
                raise ValueError(f"Handler class expected, got {target_handler!r}")

            target_ref = target  # Avoid unresolved reference 'command'
            if target_ref is None:
                target_ref = typing.cast(
                    _TargetT,
                    util.get_base_args(
                        handler,
                        base_type=self._handler_type,
                    )[0],
                )
            assert target_ref is not None

            self.subscribe(typing.cast(_TargetHandlerT, target_handler), target_ref)
            return typing.cast(_AnyHandlerT, target_handler)

        if handler is None:
            return wrap

        return wrap(handler)

    def get_handler_for(self, target: _TargetT) -> typing.Optional[_TargetHandlerT]:
        return typing.cast(
            typing.Optional[_TargetHandlerT], self._handlers.get(target, None)
        )

    def get_handlers(
        self,
    ) -> inflector.TargetHandlersViewType[_TargetT, _TargetHandlerT]:
        return types.MappingProxyType(self._handlers)

    def subscribe(
        self,
        handler: _TargetHandlerT,
        target: _TargetT,
        allow_many: bool = False,
    ) -> None:
        if not allow_many and target in self._handlers:
            raise ValueError(f"Target {target!r} is already subscribed")
        self._handlers[target] = handler

    def unsubscribe(
        self, target: _TargetT, handler: typing.Optional[_TargetHandlerT] = None
    ) -> None:
        if target not in self._handlers:
            raise ValueError(f"Command {target!r} is not subscribed")

        if handler is not None:
            if isinstance(handler, typing.Iterable):
                handlers = typing.cast(
                    typing.List[_TargetHandlerT],
                    self.get_handler_for(target) or [],
                )
                for h in handler:
                    handlers.remove(h)

                self._handlers[target] = handlers
            else:
                handlers = typing.cast(
                    typing.List[_TargetHandlerT],
                    self.get_handler_for(target) or [],
                )
                for h in handlers:
                    if h is handler:
                        handlers.remove(h)

                self._handlers[target] = handlers

            return

        self._handlers.pop(target)



from __future__ import annotations

__all__: typing.Sequence[str] = ("CommandBus",)

import abc
import typing

if typing.TYPE_CHECKING:
    from freedom.application import inflector as inflector_
    from freedom.domain import command as command_
    from freedom.domain import command_handler as handler_


class CommandBus(abc.ABC):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    async def execute(self, command: command_.Command) -> handler_.CommandResult: ...

    @abc.abstractmethod
    @typing.overload
    def listen(
        self,
        handler: typing.Literal[None],
        command: typing.Optional[command_.CommandType] = None,
    ) -> typing.Callable[
        [handler_.AnyCommandHandlerTypeT], handler_.AnyCommandHandlerTypeT
    ]: ...

    @abc.abstractmethod
    @typing.overload
    def listen(
        self,
        handler: handler_.AnyCommandHandlerTypeT,
        command: typing.Optional[command_.CommandType] = None,
    ) -> handler_.AnyCommandHandlerTypeT: ...

    @abc.abstractmethod
    def listen(
        self,
        handler: typing.Optional[handler_.AnyCommandHandlerTypeT] = None,
        command: typing.Optional[command_.CommandType] = None,
    ) -> typing.Union[
        handler_.AnyCommandHandlerTypeT,
        typing.Callable[
            [handler_.AnyCommandHandlerTypeT],
            handler_.AnyCommandHandlerTypeT,
        ],
    ]: ...

    @abc.abstractmethod
    def get_handler_for(
        self,
        command: command_.CommandType,
    ) -> typing.Optional[handler_.AnyCommandHandlerType]: ...

    @abc.abstractmethod
    def get_handlers(self) -> inflector_.TargetHandlersViewType[
        command_.CommandType,
        handler_.AnyCommandHandlerType,
    ]: ...

    @abc.abstractmethod
    def subscribe(
        self,
        handler: handler_.AnyCommandHandlerType,
        command: command_.CommandType,
    ) -> None: ...

    @abc.abstractmethod
    def unsubscribe(self, command: command_.CommandType) -> None: ...



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "EventHandler",
    "AnyEventHandlerTypeT",
    "AnyEventHandlerType",
)

import abc
import typing

import typing_extensions as typingext

AnyEventHandler: typingext.TypeAlias = "EventHandler[typing.Any]"
AnyEventHandlerType: typingext.TypeAlias = typing.Type[AnyEventHandler]
AnyEventHandlerTypeT = typing.TypeVar("AnyEventHandlerTypeT", bound=AnyEventHandlerType)

_EventT = typing.TypeVar("_EventT")


class EventHandler(abc.ABC, typing.Generic[_EventT]):
    __slots__: typing.Sequence[str] = ()

    if typing.TYPE_CHECKING:

        def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any: ...

    @abc.abstractmethod
    async def handle(self, event: _EventT, /) -> None: ...



from __future__ import annotations

import abc
import inspect
import typing

import typing_extensions as typingext

from freedom.sentinel import NOTHING
from freedom.util import get_type_hints

_KeyT = typing.TypeVar("_KeyT")
_ValueT = typing.TypeVar("_ValueT")

AnyDependencyProviderType: typingext.TypeAlias = (
    "DependencyProvider[typing.Any, typing.Any]"
)


class DependencyProvider(abc.ABC, typing.Generic[_KeyT, _ValueT]):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    def register_dependency(self, identifier: _KeyT, dependency: _ValueT) -> None: ...

    @abc.abstractmethod
    def get_dependency(self, identifier: _KeyT) -> _ValueT: ...

    def _resolve_arguments(
        self,
        callable_hints: typing.Mapping[str, typing.Any],
    ) -> typing.MutableMapping[str, typing.Any]:
        kwargs = {}
        for arg_name, arg_type in callable_hints.items():
            if arg_type is NOTHING:
                raise KeyError(f"No type hints found for param {arg_name!r}.")

            kwargs[arg_name] = self.get_dependency(arg_type)

        return kwargs

    def get_dependencies(
        self,
        callable_: typing.Callable[..., typing.Any],
    ) -> typing.Mapping[str, typing.Any]:
        sig = inspect.Signature.from_callable(callable_)
        hints = typing.get_type_hints(callable_)
        deps = {}

        for param in sig.parameters.values():
            if param.annotation is inspect.Parameter.empty:
                continue
                # raise KeyError(f"No type hints found for param {param.name!r}.")

            deps[param.name] = self.get_dependency(hints[param.name])

        return deps
        # hints = get_type_hints(callable_)
        # print(hints)
        #
        # # FIXME
        # hints.pop("self", None)
        # hints.pop("return", None)
        #
        # for arg_name, arg_type in hints.copy().items():
        #     if arg_type is NOTHING:
        #         raise KeyError(f"No type hints found for param {arg_name!r}.")
        #
        #     hints[arg_name] = self.get_dependency(arg_type)
        #
        # return hints



from __future__ import annotations

__all__: typing.Sequence[str] = ("InMemoryRepository",)

import typing

from freedom.domain import aggregate as aggregate_
from freedom.domain import entity_id
from freedom.domain import event
from freedom.domain import repository

_AggregateRootIdT = typing.TypeVar("_AggregateRootIdT")
_AggregateRootT = typing.TypeVar("_AggregateRootT", bound=aggregate_.AnyAggregateRoot)


class InMemoryRepository(repository.Repository[_AggregateRootIdT, _AggregateRootT]):
    __slots__: typing.Sequence[str] = ("_objects",)

    def __init__(self) -> None:
        self._objects: typing.Dict[entity_id.EntityId, _AggregateRootT] = {}

    async def get_by_id(
        self, aggregate_id: _AggregateRootIdT
    ) -> typing.Optional[_AggregateRootT]:
        try:
            aggregate = self._objects[aggregate_id]
        except KeyError as exc:
            raise KeyError(f"No aggregate found with id {aggregate_id!r}") from exc

        return aggregate

    async def remove_by_id(self, aggregate_id: _AggregateRootIdT) -> None:
        try:
            del self._objects[aggregate_id]
        except KeyError as exc:
            raise KeyError(f"No aggregate found with id {aggregate_id!r}") from exc

    async def remove(self, aggregate: _AggregateRootT) -> None:
        await self.remove_by_id(aggregate.id)

    async def insert(self, aggregate: _AggregateRootT) -> None:
        if aggregate.id in self._objects:
            raise ValueError(f"Aggregate with id {aggregate.id!r} already exists.")

        self._objects[aggregate.id] = aggregate

    async def save(self, aggregate: _AggregateRootT) -> None:
        self._objects[aggregate.id] = aggregate

    persist = save

    async def persist_all(self) -> None:
        return

    def collect_events(self) -> typing.List[event.Event]:
        events = []
        for aggregate in self._objects.values():
            events.extend(aggregate.collect_events())

        return events



from __future__ import annotations

__all__: typing.Sequence[str] = ("Repository",)

import abc
import typing

import typing_extensions as typingext

if typing.TYPE_CHECKING:
    from freedom.domain import event

_AggregateRootIdT = typing.TypeVar("_AggregateRootIdT")
_AggregateRootT = typing.TypeVar("_AggregateRootT")

AnyRepository: typingext.TypeAlias = "Repository[typing.Any, typing.Any]"


class Repository(abc.ABC, typing.Generic[_AggregateRootIdT, _AggregateRootT]):
    __slots__: typing.Sequence[str] = ()

    @abc.abstractmethod
    async def get_by_id(
        self, aggregate_id: _AggregateRootIdT
    ) -> typing.Optional[_AggregateRootT]: ...

    @abc.abstractmethod
    async def remove_by_id(self, aggregate_id: _AggregateRootIdT) -> None: ...

    @abc.abstractmethod
    async def remove(self, aggregate: _AggregateRootT) -> None: ...

    @abc.abstractmethod
    async def insert(self, aggregate: _AggregateRootT) -> None: ...

    @abc.abstractmethod
    async def save(self, aggregate: _AggregateRootT) -> None: ...

    @abc.abstractmethod
    async def persist(self, aggregate: _AggregateRootT) -> None: ...

    @abc.abstractmethod
    async def persist_all(self) -> None: ...

    @abc.abstractmethod
    def collect_events(self) -> typing.List[event.Event]: ...



from __future__ import annotations

import abc
import types
import typing

import typing_extensions as typingext

_TargetHandlerT = typing.TypeVar("_TargetHandlerT")
_TargetT = typing.TypeVar("_TargetT")

TargetHandlersType: typingext.TypeAlias = (
    typing.MutableMapping
)  # [_TargetT, _TargetHandlerT]
TargetHandlersViewType: typingext.TypeAlias = (
    types.MappingProxyType
)  # [_TargetT, _TargetHandlerT]

_AnyHandlerT = typing.TypeVar("_AnyHandlerT")


class Inflector(abc.ABC, typing.Generic[_TargetT, _TargetHandlerT]):
    __slots__: typing.Sequence[str] = ()

    # @abc.abstractmethod
    # @typing.overload
    # def listen(
    #     self,
    #     handler: typing.Literal[None],
    #     target: typing.Optional[_TargetT] = None,
    # ) -> typing.Callable[[_TargetHandlerT], _TargetHandlerT]: ...
    #
    # @abc.abstractmethod
    # @typing.overload
    # def listen(
    #     self,
    #     handler: _TargetHandlerT,
    #     target: typing.Optional[_TargetT] = None,
    # ) -> _TargetHandlerT: ...

    @abc.abstractmethod
    def listen(
        self,
        handler: typing.Optional[_AnyHandlerT] = None,
        target: typing.Optional[_TargetT] = None,
    ) -> typing.Union[_AnyHandlerT, typing.Callable[[_AnyHandlerT], _AnyHandlerT]]: ...

    @abc.abstractmethod
    def get_handler_for(self, target: _TargetT) -> typing.Optional[_TargetHandlerT]: ...

    @abc.abstractmethod
    def get_handlers(self) -> TargetHandlersViewType[_TargetT, _TargetHandlerT]: ...

    @abc.abstractmethod
    def subscribe(
        self,
        handler: _TargetHandlerT,
        target: _TargetT,
        allow_many: bool = False,
    ) -> None: ...

    @abc.abstractmethod
    def unsubscribe(
        self,
        target: _TargetT,
        handler: typing.Optional[_TargetHandlerT] = None,
    ) -> None: ...



from __future__ import annotations

import typing

import typing_extensions as typingext

from freedom.domain import entity
from freedom.domain import entity_id
from freedom.domain import event as event_

_AggregateIdT = typing.TypeVar("_AggregateIdT", bound=entity_id.EntityId)

AnyAggregateRoot: typingext.TypeAlias = "AggregateRoot[typing.Any]"


class AggregateRoot(entity.Entity[_AggregateIdT]):
    __slots__: typing.Sequence[str] = ("_uncommitted_events",)

    def __init__(self, id: _AggregateIdT) -> None:
        super().__init__(id=id)
        self._uncommitted_events: typing.List[event_.Event] = []

    @property
    def uncommitted_events(self) -> typing.List[event_.Event]:
        return self._uncommitted_events

    def collect_events(self) -> typing.Sequence[event_.Event]:
        events = self._uncommitted_events[:]
        self._uncommitted_events.clear()
        return events

    def record_that(self, event: event_.Event) -> None:
        self._uncommitted_events.append(event)



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "LoggerMiddleware",
    "MiddlewareChain",
)

import logging
import typing

from freedom import util
from freedom.application import middleware as middleware_


class LoggerMiddleware(middleware_.Middleware):
    __slots__: typing.Sequence[str] = (
        "_logger",
        "_received_level",
        "_succeeded_level",
        "_failed_level",
    )

    def __init__(
        self,
        logger: logging.Logger,
        *,
        received_level: int = logging.DEBUG,
        succeeded_level: int = logging.DEBUG,
        failed_level: int = logging.ERROR,
    ) -> None:
        self._logger = logger
        self._received_level = received_level
        self._succeeded_level = succeeded_level
        self._failed_level = failed_level

    async def __call__(
        self,
        message: typing.Any,
        next_middleware: middleware_.PartialMiddlewareSigType,
    ) -> typing.Any:
        message_type = type(message)

        self._logger.log(
            self._received_level,
            f"Message received: ${message_type}",
        )

        try:
            result = await next_middleware(message)
        except Exception as exc:
            self._logger.log(
                self._failed_level,
                f"Message failed: ${message_type}",
                exc_info=True,
            )
            raise exc

        self._logger.log(
            self._succeeded_level,
            f"Message succeeded: ${message_type}",
        )
        return result


class MiddlewareChain(middleware_.MiddlewareChain):
    def __init__(
        self,
        executor_middleware: middleware_.MiddlewareSigType,
        middlewares: typing.Optional[typing.List[middleware_.MiddlewareSigType]] = None,
    ) -> None:
        self.middlewares = middlewares or []
        self.executor_middleware = executor_middleware

    async def __call__(self, message: typing.Any) -> typing.Any:
        all_middlewares = (self.middlewares or []) + [self.executor_middleware]

        chain = typing.cast(middleware_.MiddlewareSigType, util.empty_coro)
        for middleware in reversed(all_middlewares):
            chain = self.set_next(middleware, chain)

        result = await chain(message)
        return result

    def set_next(
        self,
        middleware: middleware_.MiddlewareSigType,
        next_middleware: middleware_.MiddlewareSigType,
    ) -> middleware_.PartialMiddlewareSigType:
        async def wrap(message: typing.Any) -> typing.Any:
            result = await middleware(message, next_middleware)
            return result

        return wrap



from __future__ import annotations

import abc
import asyncio
import inspect
import sys
import typing

from freedom import sentinel

FROZEN_STR: typing.Final[str] = "__frozen__"
UNFREEZE_ATTRS_STR: typing.Final[str] = "__unfreeze_attrs__"

_T = typing.TypeVar("_T")


def __frozen_setattr__(
    self: typing.Any,
    key: typing.Any,
    value: typing.Any,
) -> None:
    if (
        self.__frozen__
        and sys._getframe().f_back.f_code.co_name not in self.__unfreeze_attrs__
    ):
        raise ImmutableObjectError("Object is immutable.")

    if hasattr(self, "__dict__"):
        self.__dict__[key] = value
        return

    object.__setattr__(self, key, value)


def __frozen_delattr__(self: typing.Any, item: typing.Any) -> None:
    if (
        self.__frozen__
        and sys._getframe().f_back.f_code.co_name not in self.__unfreeze_attrs__
    ):
        raise ImmutableObjectError("Object is immutable.")

    object.__delattr__(self, item)


async def empty_coro(*_: typing.Any, **__: typing.Any) -> typing.Any:
    pass


def get_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.get_event_loop_policy().get_event_loop()

    return loop


def as_completed_future(
    result: typing.Optional[_T] = None,
) -> asyncio.Future[typing.Optional[_T]]:
    future = get_loop().create_future()
    future.set_result(result)
    return future


def get_type_hints(
    obj: typing.Any,
    *,
    subclass_of: typing.Optional[typing.Tuple[typing.Type[typing.Any], ...]] = None,
) -> typing.Dict[str, typing.Type[typing.Any]]:
    hints = {
        k: (
            v.annotation
            if v.annotation is not inspect.Parameter.empty
            else sentinel.NOTHING
        )
        for k, v in inspect.Signature.from_callable(obj).parameters.items()
    }
    hints.update(typing.get_type_hints(obj))

    if subclass_of is not None:
        hints = {
            k: v
            for k, v in hints.items()
            if inspect.isclass(v) and issubclass(v, subclass_of)
        }

    return hints


def unpack_unions(
    hints: typing.Dict[str, typing.Type[typing.Any]],
) -> typing.Dict[str, typing.Type[typing.Any]]:
    unions = {}

    def _unpack_item(
        key: str,
        union_value: typing.Any,
        mapping: typing.Dict[str, typing.List[typing.Type[typing.Any]]],
    ) -> None:
        if key not in mapping:
            mapping[key] = []

        args = typing.get_args(union_value)
        for arg in args:
            if typing.get_origin(arg) is typing.Union:
                _unpack_item(key, arg, mapping)

            mapping[key].append(arg)

    for k, v in hints.items():
        if typing.get_origin(v) is typing.Union:
            _unpack_item(k, v, unions)

    hints.update(unions)
    return hints


def get_base_args(
    obj: typing.Any,
    base_type: typing.Type[typing.Any],
) -> typing.Tuple[typing.Type[typing.Any], ...]:
    bases = typing.cast(
        typing.Union[typing.Tuple[typing.Type[typing.Any], ...], None],
        getattr(obj, "__orig_bases__", None),
    )
    if bases is None:
        raise AttributeError(f"Cant find orig bases in obj {obj!r}")

    for base in bases:
        if (orig := typing.get_origin(base)) is None:
            # get_base_arg can handle only parametrized generics
            continue

        if orig is base_type:
            args = getattr(base, "__args__", ())
            return args

    raise ValueError(f"Cant find base that matches {base_type!r}")


class ImmutableObjectError(Exception):
    pass


class ImmutableMeta(abc.ABCMeta):
    def __new__(
        mcs: typing.Type[ImmutableMeta],
        name: str,
        bases: typing.Tuple[typing.Type[typing.Any], ...],
        attrs: typing.Dict[str, typing.Any],
    ) -> typing.Any:
        frozen = attrs.pop(FROZEN_STR, True)
        unfreeze = attrs.pop(UNFREEZE_ATTRS_STR, ())
        unfreeze += ("__frozen_setattr__", "__frozen_delattr__", "__init__")

        #
        attrs[UNFREEZE_ATTRS_STR] = unfreeze
        attrs[FROZEN_STR] = frozen

        attrs["__setattr__"] = attrs["__setitem__"] = __frozen_setattr__
        attrs["__delattr__"] = attrs["__delitem__"] = __frozen_delattr__
        return super().__new__(mcs, name, bases, attrs)



from __future__ import annotations

__all__: typing.Sequence[str] = ("Command", "CommandType")

import typing

from src.domain.valueobject import ValueObject

CommandType = typing.Type["Command"]


class Command(ValueObject):
    __slots__: typing.Sequence[str] = ()



from __future__ import annotations

__all__: typing.Sequence[str] = ("InMemoryCommandBus",)

import typing

from freedom.application import application
from freedom.application import command_bus
from freedom.application import inflector as inflector_
from freedom.domain import command as command_
from freedom.domain import command_handler as handler_
from freedom.infrastructure import inflector as inflector_impl
from freedom.infrastructure import middleware as middleware_impl

if typing.TYPE_CHECKING:
    from freedom.application import middleware
    from freedom.application import provider as provider_


class InMemoryCommandBus(command_bus.CommandBus):
    __slots__: typing.Sequence[str] = (
        "_handlers",
        "_provider",
        "_inflector",
        "_middlewares_chain",
    )

    def __init__(
        self,
        middlewares: typing.Optional[typing.List[middleware.MiddlewareSigType]] = None,
        provider: typing.Optional[provider_.AnyDependencyProviderType] = None,
        inflector: typing.Optional[
            inflector_.Inflector[command_.CommandType, handler_.AnyCommandHandlerType]
        ] = None,
    ) -> None:
        if inflector is None:
            inflector = typing.cast(
                inflector_.Inflector[
                    command_.CommandType,
                    handler_.AnyCommandHandlerType,
                ],
                inflector_impl.InMemoryInflector(
                    handler_type=handler_.CommandHandler,
                ),
            )

        self._middlewares_chain = middleware_impl.MiddlewareChain(
            self._executor_middleware,
            middlewares or [],
        )
        self._inflector = inflector
        self._provider = provider
        self._handlers: inflector_.TargetHandlersType[
            command_.CommandType, handler_.AnyCommandHandlerType
        ] = {}

    @typing.overload
    def listen(
        self,
        handler: typing.Literal[None],
        command: typing.Optional[command_.CommandType] = None,
    ) -> typing.Callable[
        [handler_.AnyCommandHandlerTypeT], handler_.AnyCommandHandlerTypeT
    ]: ...

    @typing.overload
    def listen(
        self,
        handler: handler_.AnyCommandHandlerTypeT,
        command: typing.Optional[command_.CommandType] = None,
    ) -> handler_.AnyCommandHandlerTypeT: ...

    def listen(
        self,
        handler: typing.Optional[handler_.AnyCommandHandlerTypeT] = None,
        command: typing.Optional[command_.CommandType] = None,
    ) -> typing.Union[
        handler_.AnyCommandHandlerTypeT,
        typing.Callable[
            [handler_.AnyCommandHandlerTypeT],
            handler_.AnyCommandHandlerTypeT,
        ],
    ]:
        return self._inflector.listen(handler, command)

    def get_handler_for(
        self, command: command_.CommandType
    ) -> typing.Optional[handler_.AnyCommandHandlerType]:
        return self._inflector.get_handler_for(command)

    def get_handlers(
        self,
    ) -> inflector_.TargetHandlersViewType[
        command_.CommandType, handler_.AnyCommandHandlerType
    ]:
        return self._inflector.get_handlers()

    def subscribe(
        self,
        handler: handler_.AnyCommandHandlerType,
        command: command_.CommandType,
    ) -> None:
        self._inflector.subscribe(handler, command)

    def unsubscribe(self, command: command_.CommandType) -> None:
        self._inflector.unsubscribe(command)

    async def execute(self, command: command_.Command) -> handler_.CommandResult:
        result = typing.cast(
            handler_.CommandResult,
            await self._middlewares_chain(command),
        )
        return result

    @typing.no_type_check
    async def _executor_middleware(
        self,
        command: command_.Command,
        _: middleware.MiddlewareSigType,
    ) -> handler_.CommandResult:
        command_type = type(command)
        handler = self.get_handler_for(command_type)
        if handler is not None:
            kwargs: typing.Dict[str, typing.Any] = {}

            if self._provider is not None:
                kwargs.update(self._provider.get_dependencies(handler.__init__))

            handler = handler(**kwargs)
            result = typing.cast(
                handler_.CommandResult,
                application.collect_domain_events(
                    await handler.handle(command),
                    kwargs,
                ),
            )
            return result

        return handler_.CommandResult.failure(
            fail_message="Handler not found.",
        )



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "AnyCommandHandlerTypeT",
    "CommandHandler",
    "CommandResultError",
    "CommandResult",
)

import abc
import sys
import typing

import typing_extensions as typingext

from freedom.domain.command import Command

if typing.TYPE_CHECKING:
    from freedom import type_hints
    from freedom.domain import entity_id as entity_id_
    from freedom.domain import event as event_

AnyCommandHandler: typingext.TypeAlias = "CommandHandler[typing.Any]"
AnyCommandHandlerType: typingext.TypeAlias = typing.Type[AnyCommandHandler]
AnyCommandHandlerTypeT = typing.TypeVar(
    "AnyCommandHandlerTypeT", bound=AnyCommandHandlerType
)

_CommandT = typing.TypeVar("_CommandT", bound=Command)


class CommandHandler(abc.ABC, typing.Generic[_CommandT]):
    __slots__: typing.Sequence[str] = ()

    if typing.TYPE_CHECKING:

        def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any: ...

    @abc.abstractmethod
    async def handle(self, command: _CommandT, /) -> CommandResult: ...


class CommandResultError:
    __slots__: typing.Sequence[str] = (
        "message",
        "exception",
        "exception_info",
    )

    def __init__(
        self,
        message: str,
        exception: typing.Optional[BaseException] = None,
        exception_info: typing.Optional[type_hints.OptionalSysExcInfo] = None,
    ) -> None:
        self.message = message
        self.exception = exception
        self.exception_info = exception_info


class CommandResult:
    __slots__: typing.Sequence[str] = (
        "entity_id",
        "events",
        "errors",
    )

    def __init__(
        self,
        entity_id: typing.Optional[entity_id_.EntityId] = None,
        events: typing.List[event_.Event] = None,
        errors: typing.List[CommandResultError] = None,
    ) -> None:
        self.entity_id = entity_id
        self.events = events or []
        self.errors = errors or []

    def has_errors(self) -> bool:
        return len(self.errors) > 0

    def is_success(self) -> bool:
        return not self.has_errors()

    def add_error(self, command_error: CommandResultError, /) -> None:
        self.errors.append(command_error)

    @classmethod
    def failure(
        cls,
        fail_message: str,
        exception: typing.Optional[BaseException] = None,
        exception_info: typing.Optional[type_hints.OptionalSysExcInfo] = None,
    ) -> CommandResult:
        if exception_info is None:
            exception_info = sys.exc_info()

        result_error = CommandResultError(
            message=fail_message,
            exception=exception,
            exception_info=exception_info,
        )
        result = cls()
        result.add_error(result_error)
        return result

    @classmethod
    def success(
        cls,
        entity_id: typing.Optional[entity_id_.EntityId] = None,
        event: typing.Optional[event_.Event] = None,
        events: typing.Optional[typing.Sequence[event_.Event]] = None,
    ) -> CommandResult:
        events = events or []
        if event is not None:
            events.append(event)

        result = cls(
            entity_id=entity_id,
            events=events,
        )
        return result



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "SysExcInfo",
    "OptionalSysExcInfo",
)

import types
import typing

import typing_extensions as typingext

SysExcInfo: typingext.TypeAlias = typing.Tuple[
    typing.Type[BaseException], BaseException, types.TracebackType
]
OptionalSysExcInfo: typingext.TypeAlias = typing.Union[
    SysExcInfo, typing.Tuple[None, None, None]
]



from __future__ import annotations

__all__: typing.Sequence[str] = (
    "NOTHING",
    "NothingOr",
)

import typing

import typing_extensions

_SelfT = typing.TypeVar("_SelfT")


def __new__(_: typing.Type[_SelfT]) -> _SelfT:
    raise TypeError(f"Cannot create multiple instances of NothingType.")


class NothingType:
    __slots__: typing.Sequence[str] = ()

    def __str__(self) -> str:
        return "NOTHING"

    def __repr__(self) -> str:
        return "NOTHING"

    def __reduce__(self) -> str:
        return "NOTHING"

    def __getstate__(self) -> typing.Any:
        return False

    def __bool__(self) -> typing.Literal[False]:
        return False

    def __copy__(self) -> typing_extensions.Self:
        return self

    def __deepcopy__(
        self, memo: typing.MutableMapping[int, typing.Any]
    ) -> typing_extensions.Self:
        memo[id(self)] = self
        return self


NOTHING = NothingType()
NothingType.__new__ = __new__
del __new__

_T = typing.TypeVar("_T")
NothingOr = typing.Union[_T, NothingType]



from __future__ import annotations

import abc
import typing

from freedom.domain.valueobject import ValueObject


class BusinessRule(ValueObject, abc.ABC):
    @abc.abstractmethod
    def is_broken(self) -> bool: ...

    @abc.abstractmethod
    def render_broken_rule(self) -> str: ...


class BusinessRuleValidationMixin:
    __slots__: typing.Sequence[str] = ()

    def check_rules(self, *rules: BusinessRule) -> None:
        violated_rules = []
        for rule in rules:
            if rule.is_broken():
                violated_rules.append(rule)

        if violated_rules:
            raise BusinessRuleValidationError(*violated_rules)


class BusinessRuleValidationError(Exception):
    __slots__: typing.Sequence[str] = ("_rules",)

    def __init__(self, *rules: BusinessRule) -> None:
        self._rules = rules
        super().__init__(self._render_broken_rules())

    def _render_broken_rules(self) -> str:
        broken_rules = ",".join(type(rule).__name__ for rule in self._rules)
        return f"Rules {broken_rules!r} are broken."



from __future__ import annotations

import typing

from freedom import util


class ValueObject(metaclass=util.ImmutableMeta):
    @classmethod
    def unfreeze(cls) -> None:
        setattr(cls, util.FROZEN_STR, False)

    @classmethod
    def freeze(cls) -> None:
        setattr(cls, util.FROZEN_STR, True)

    @classmethod
    def is_frozen(cls) -> bool:
        is_frozen = typing.cast(bool, getattr(cls, util.FROZEN_STR, False))
        return is_frozen
